import pathlib
import pickle
import time

from flexcache import DiskCacheByMTime

# These sleep time is needed when run on GitHub Actions
# If not given or too short, some mtime changes are not visible.
FS_SLEEP = 0.010


def parser(p: pathlib.Path):
    return p.read_bytes()


def test_new_cache_date(tmp_path):
    # Generate a definition file
    dfile = tmp_path / "definitions.txt"
    dfile.write_bytes(b"1234")

    dc = DiskCacheByMTime(tmp_path)
    content = dc.load(dfile)[0]

    time.sleep(FS_SLEEP)

    assert len(tuple(tmp_path.glob("*.pickle"))) == 0
    assert len(tuple(tmp_path.glob("*.json"))) == 0

    # First, the cache should be missed
    assert content is None
    dc.save(pickle.dumps(dfile.read_bytes()), dfile)

    # There should be a cache file now
    assert len(tuple(tmp_path.glob("*.pickle"))) == 1
    assert len(tuple(tmp_path.glob("*.json"))) == 1

    content = pickle.loads(dc.load(dfile)[0])
    # Now, the cache should be hit
    assert content == b"1234"

    # Modify the definition file
    # Add some sleep to make sure that the time stamp difference is significant.
    time.sleep(FS_SLEEP)
    dfile.write_bytes(b"1235")
    time.sleep(FS_SLEEP)

    # Modify the time stamp of the old cache files so they are newer
    for p in tmp_path.glob("*.pickle"):
        p.touch()

    # Verify that the cached file (not the definition file), was loaded.
    d = pickle.loads(dc.load(dfile)[0])
    assert d == b"1234"


def test_file_changed(tmp_path):
    # Generate a definition file
    dfile = tmp_path / "definitions.txt"
    dfile.write_bytes(b"1234")

    dc = DiskCacheByMTime(tmp_path)
    content = dc.load(dfile)[0]

    assert len(tuple(tmp_path.glob("*.pickle"))) == 0
    assert len(tuple(tmp_path.glob("*.json"))) == 0

    time.sleep(FS_SLEEP)

    # First, the cache should be missed
    assert content is None
    dc.save(pickle.dumps(dfile.read_bytes()), dfile)

    # There should be a cache file now
    assert len(tuple(tmp_path.glob("*.pickle"))) == 1
    assert len(tuple(tmp_path.glob("*.json"))) == 1

    content = pickle.loads(dc.load(dfile)[0])
    # Now, the cache should be hit
    assert content == b"1234"

    # Modify the definition file
    # Add some sleep to make sure that the time stamp difference is significant.
    time.sleep(FS_SLEEP)
    dfile.write_bytes(b"1235")

    # Verify that the cached was not loaded
    content = dc.load(dfile)[0]
    assert content is None


def test_func(tmp_path):
    # Generate a definition file
    dfile = tmp_path / "definitions.txt"
    dfile.write_bytes(b"1234")

    dc = DiskCacheByMTime(tmp_path)
    assert dc.load(dfile, converter=parser)[0] == b"1234"
    # There should be a cache file now
    assert len(tuple(tmp_path.glob("*.pickle"))) == 1
    assert len(tuple(tmp_path.glob("*.json"))) == 1

    # Modify the definition file
    # Add some sleep to make sure that the time stamp difference is significant.
    dfile.write_bytes(b"1235")

    # Verify that the cache was not loaded as the content of the original file
    # has changed.
    assert dc.load(dfile, converter=parser)[0] == b"1235"
    # There should be still ONE cache files now
    assert len(tuple(tmp_path.glob("*.pickle"))) == 1
    assert len(tuple(tmp_path.glob("*.json"))) == 1



"""
    flexcache
    ~~~~~~~~~

    Classes for persistent caching and invalidating cached objects,
    which are built from a source object and a (potentially expensive)
    conversion function.

    :copyright: 2022 by flexcache Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from __future__ import annotations

from importlib.metadata import version

try:  # pragma: no cover
    __version__ = version("flexcache")
except Exception:  # pragma: no cover
    # we seem to have a local copy not installed without setuptools
    # so the reported version will be unknown
    __version__ = "unknown"

from .flexcache import (
    BaseHeader,
    BasicPythonHeader,
    DiskCache,
    DiskCacheByHash,
    DiskCacheByMTime,
    InvalidateByExist,
    InvalidateByMultiPathsMtime,
    InvalidateByPathMTime,
    NameByFields,
    NameByFileContent,
    NameByHashIter,
    NameByMultiPaths,
    NameByObj,
    NameByPath,
)

__all__ = (
    "__version__",
    "BaseHeader",
    "BasicPythonHeader",
    "NameByFields",
    "NameByFileContent",
    "NameByObj",
    "NameByPath",
    "NameByMultiPaths",
    "NameByHashIter",
    "DiskCache",
    "DiskCacheByHash",
    "DiskCacheByMTime",
    "InvalidateByExist",
    "InvalidateByPathMTime",
    "InvalidateByMultiPathsMtime",
)



import json
import pathlib
import pickle
import time
from dataclasses import asdict as dc_asdict
from dataclasses import dataclass

import pytest

import flexcache

# These sleep time is needed when run on GitHub Actions
# If not given or too short, some mtime changes are not visible.
FS_SLEEP = 0.010


def test_empty(tmp_path):
    @dataclass(frozen=True)
    class Hdr(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BaseHeader
    ):
        pass

    hdr = Hdr("123", "myconverter")
    assert tuple(hdr.for_cache_name()) == ("myconverter".encode("utf-8"),)

    p1 = tmp_path / "cache.pickle"
    assert not hdr.is_valid(p1)
    p1.touch()
    assert hdr.is_valid(p1)
    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_basic_python():
    @dataclass(frozen=True)
    class Hdr(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BasicPythonHeader
    ):
        pass

    hdr = Hdr("123", "myconverter")
    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 4

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_name_by_content(tmp_path):
    @dataclass(frozen=True)
    class Hdr(
        flexcache.InvalidateByExist, flexcache.NameByFileContent, flexcache.BaseHeader
    ):
        pass

    p = tmp_path / "source.txt"
    p.write_bytes(b"1234")
    hdr = Hdr(p, "myconverter")

    assert hdr.source_path == p
    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 2
    assert cn[1] == b"1234"

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_name_by_path(tmp_path):
    @dataclass(frozen=True)
    class Hdr(
        flexcache.NameByPath, flexcache.InvalidateByPathMTime, flexcache.BaseHeader
    ):
        pass

    p = tmp_path / "source.txt"
    p.write_bytes(b"1234")
    hdr = Hdr(p, "myconverter")

    assert hdr.source_path == p
    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 2
    assert cn[1] == bytes(p.resolve())

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_name_by_paths(tmp_path):
    @dataclass(frozen=True)
    class Hdr(
        flexcache.NameByMultiPaths,
        flexcache.InvalidateByMultiPathsMtime,
        flexcache.BaseHeader,
    ):
        pass

    p0 = tmp_path / "source0.txt"
    p0.touch()

    time.sleep(FS_SLEEP)

    p1 = tmp_path / "source1.txt"
    p2 = tmp_path / "source2.txt"
    p1.write_bytes(b"1234")
    p2.write_bytes(b"1234")
    hdr = Hdr((p1, p2), "myconverter")

    time.sleep(FS_SLEEP)

    p3 = tmp_path / "source3.txt"
    p3.touch()

    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 3
    assert cn[1] == bytes(p1.resolve())
    assert cn[2] == bytes(p2.resolve())

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False

    assert not hdr.is_valid(tmp_path / "not.txt")
    assert not hdr.is_valid(p0)
    assert hdr.is_valid(p3)

    hdr = Hdr.from_strings((str(p1), str(p2)), "myconverter")
    assert hdr.source_paths == (p1, p2)


def test_name_by_obj(tmp_path):
    @dataclass(frozen=True)
    class Hdr(flexcache.InvalidateByExist, flexcache.NameByObj, flexcache.BaseHeader):
        pass

    hdr = Hdr((1, 2, 3), "myconverter")

    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 2
    assert hdr.pickle_protocol == pickle.HIGHEST_PROTOCOL
    assert cn[1] == pickle.dumps((1, 2, 3), hdr.pickle_protocol)

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_name_by_hash(tmp_path):
    @dataclass(frozen=True)
    class Hdr(
        flexcache.InvalidateByExist, flexcache.NameByHashIter, flexcache.BaseHeader
    ):
        pass

    hdr = Hdr(("b", "a", "c"), "myconverter")

    cn = tuple(hdr.for_cache_name())
    assert len(cn) == 4
    assert cn[1] == b"a"
    assert cn[2] == b"b"
    assert cn[3] == b"c"

    try:
        json.dumps({k: str(v) for k, v in dc_asdict(hdr).items()})
    except Exception:
        assert False


def test_predefined_headers(tmp_path):
    fn = "source.txt"
    hdr0 = flexcache.DiskCacheByMTime.Header.from_string(fn, "123")
    assert isinstance(hdr0.source_path, pathlib.Path)
    assert str(hdr0.source_path) == fn

    hdr1 = flexcache.DiskCacheByHash.Header.from_string(fn, "123")
    assert isinstance(hdr1.source_path, pathlib.Path)
    assert str(hdr1.source_path) == fn


def test_wrong_type():
    @dataclass(frozen=True)
    class Hdr(flexcache.NameByPath, flexcache.InvalidateByPathMTime):
        pass

    with pytest.raises(TypeError):
        Hdr("testing", "my_converter")



import pytest

import flexcache
from flexcache import DiskCache


def test_register(tmp_path):
    c = DiskCache(tmp_path)

    class Header(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BaseHeader
    ):
        @classmethod
        def from_int(cls, source, converter_id):
            return cls(bytes(source), converter_id)

    c.register_header_class(int, Header.from_int)

    c.load(3)

    with pytest.raises(TypeError):
        c.load(3j)


def test_missing_cache_path(tmp_path):
    c = DiskCache(tmp_path)

    class Header(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BaseHeader
    ):
        @classmethod
        def from_int(cls, source, converter_id):
            return cls(bytes(source), converter_id)

    hdr = Header("123", "456")
    assert c.rawsave(hdr, "789").stem == c.cache_stem_for(hdr)
    assert c.rawload(hdr) == "789"


def test_converter_id(tmp_path):
    c = DiskCache(tmp_path)

    class Header(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BaseHeader
    ):
        @classmethod
        def from_int(cls, source, converter_id):
            return cls(bytes(source), converter_id)

    c.register_header_class(int, Header.from_int)

    def func(n):
        return n * 2

    content, this_hash = c.load(21, func)
    assert content == 42
    assert c.load(21, "func") == (content, this_hash)
    assert c.save(content, 21, "func") == this_hash


def test_converter_pass_hash(tmp_path):
    c = DiskCache(tmp_path)

    class Header(
        flexcache.InvalidateByExist, flexcache.NameByFields, flexcache.BaseHeader
    ):
        @classmethod
        def from_int(cls, source, converter_id):
            return cls(bytes(source), converter_id)

    c.register_header_class(int, Header.from_int)

    def func(n, a_hash):
        return (n, a_hash)

    content, this_hash = c.load(21, func, True)
    assert content == (21, this_hash)
    assert c.load(21, "func", True) == (content, this_hash)
    assert c.save(content, 21, "func") == this_hash
